import os
import hashlib
import platform
from typing import List, Dict, Any  # noqa: F401
from .common import *
import distro


# Defaults
ARCHITECTURE = platform.machine()
OS = platform.system()
DISPLAY_NAME = ""
CAPABILITY_NAME = ""
AUTHOR = ""
EMAIL = ""
DESCRIPTION = ""
HOMEPAGE = ""
LICENSE = ""
MODULE_NAME = ""
MODULE_VERSION = "1.0"
MODULE_SEMANTIC_VERSION = '0.0.1'
MODULE_COMMANDS = []  # type: List[Dict[str, str]]
EXCLUDE_COMMANDS = []  # type: List[str]
MODULE_CAPABILITIES = []  # type: List[Dict[str, str]]
MODULE_DEPENDENCIES = []  # type: Dict[str, Dict[str, str]]
MODULE_OPTIONAL_DEPENDENCIES = []  # type: Dict[str, Dict[str, str]]
COMMAND_LINE_ARGS = ""
RUN_COMMAND_LINE_ARGS = None
MIN_REDIS_VERSION = "4.0"
MIN_REDIS_PACK_VERSION = "5.2"
COMPATIBLE_REDIS_VERSION = "7.4"
BIGSTORE_VERSION_2_SUPPORT = False
RAMP_FORMAT_VERSION = 1
CONFIG_COMMAND = ""
OVERIDE_COMMAND = []  # type: List[Dict[str, str]]
ADD_COMMAND = []  # type: List[Dict[str, str]]
CRDB_ARGS = {}  # type: List[Dict[str, str]]
REDIS_ARGS = {}  # type: List[Dict[str, str]]

FIELDS = [
    "architecture",
    "author",
    "capabilities",
    "command_line_args",
    "run_command_line_args",
    "commands",
    "config_command",
    "dependencies",
    "optional-dependencies",
    "description",
    "display_name",
    "capability_name",
    "email",
    "exclude_commands",
    "git_sha",
    "homepage",
    "license",
    "min_redis_pack_version",
    "min_redis_version",
    "compatible_redis_version",
    "bigstore_version_2_support",
    "module_file",
    "module_name",
    "os",
    "overide_command",
    "add_command",
    "ramp_format_version",
    "semantic_version",
    "sha256",
    "version",
    "crdb",
    "redis_args",
    "operating_systems"
]  # type: List[str]


def sha256_checksum(filename, block_size=65536):
    # type: (str, int) -> str
    """Computes sha256 for given file"""
    sha256 = hashlib.sha256()
    with open(filename, 'rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha256.update(block)
    return sha256.hexdigest()

RLEC_OS_MAP = {
    "rhel7": "rhel7",
    "rhel8": "rhel8",
    "centos7": "rhel7",
    "centos8": "rhel8",
    "rocky8": "rhel8",
    "almalinux8": "rhel8",
    "oracle8": "rhel8",
    "rocky9": "rhel9"
}

def get_curr_os():
    global RLEC_OS_MAP
    curr_os = '%s%s' % (distro.id(), distro.version_parts()[0])
    rlec_os = RLEC_OS_MAP.get(curr_os, curr_os)
    return rlec_os

def create_default_metadata(module_path):
    # type: (str) -> Dict[str, Any]
    """Creates a default metadata"""
    return {
        "module_name": MODULE_NAME,
        "module_file": os.path.basename(module_path),
        "os": OS,
        "architecture": ARCHITECTURE,
        "version": MODULE_VERSION,
        "semantic_version": MODULE_SEMANTIC_VERSION,
        "display_name": DISPLAY_NAME,
        "capability_name": CAPABILITY_NAME,
        "author": AUTHOR,
        "email": EMAIL,
        "description": DESCRIPTION,
        "homepage": HOMEPAGE,
        "license": LICENSE,
        "command_line_args": COMMAND_LINE_ARGS,
        "run_command_line_args": RUN_COMMAND_LINE_ARGS,
        "capabilities": MODULE_CAPABILITIES,
        "dependencies": MODULE_DEPENDENCIES,
        "optional-dependencies": MODULE_OPTIONAL_DEPENDENCIES,
        "min_redis_version": MIN_REDIS_VERSION,
        "min_redis_pack_version": MIN_REDIS_PACK_VERSION,
        "compatible_redis_version": COMPATIBLE_REDIS_VERSION,
        "bigstore_version_2_support": BIGSTORE_VERSION_2_SUPPORT,
        "sha256": sha256_checksum(module_path),
        "commands": MODULE_COMMANDS,
        "ramp_format_version": RAMP_FORMAT_VERSION,
        "config_command": CONFIG_COMMAND,
        "exclude_commands": EXCLUDE_COMMANDS,
        "overide_command": OVERIDE_COMMAND,
        "add_command": ADD_COMMAND,
        "crdb": CRDB_ARGS,
        "redis_args": REDIS_ARGS,
        "operating_systems": [get_curr_os()],
    }



import os.path

from RAMP.disposableredis import DisposableRedis
from RAMP import config
from .common import *

OK = "OK"

class Module(object):
    def __init__(self, module_name, module_version):
        self.name = module_name
        self.version = module_version
        self.commands = []

    def add_command(self, command):
        self.commands.append(command)

class ModuleCommand(object):
    def __init__(self, command_name, command_arity, flags, first_key, last_key, step):
        self.command_name = command_name
        self.command_arity = command_arity
        self.flags = flags
        self.first_key = first_key
        self.last_key = last_key
        self.step = step

    def to_dict(self):
        return self.__dict__

def redis(extra_args=None):
    redis_client = DisposableRedis(verbose=config.debug, **extra_args)
    return redis_client

def _get_modules_list(redis_client):
    """
    Finds out which modules are loaded into Redis.
    """
    modules_list = redis_client.module_list()
    # MODULE LIST response contains an array of module name and version.
    #    1) 1) "name"
    #       2) "graph"
    #       3) "ver"
    #       4) (integer) 1

    loaded_modules = []
    loaded_modules = [(m['name'], float(m['ver'])) for m in modules_list]

    return loaded_modules

def _load_module(redis_client, path_to_module, module_args):
    """
    Loads given module to redis.
    :redis_client: open connection to redis
    :param path_to_module: where does the module file is located.
    Assuming only a single module is loaded.
    """
    previously_loaded_modules = _get_modules_list(redis_client)
    resp = redis_client.execute_command("MODULE LOAD {} {}".format(
        os.path.abspath(path_to_module), module_args))
    if resp != OK:
        return None

    loaded_modules = _get_modules_list(redis_client)
    relevant_module_list = [x for x in loaded_modules if x not in previously_loaded_modules]
    if len(relevant_module_list) != 1:
        return None

    module_name, module_version = relevant_module_list[0]
    return Module(module_name, module_version)

def _get_redis_commands(redis_client):
    """
    Retrieves a set of commands from Redis
    """
    commands = redis_client.command()
    return commands

def _get_redis_command_info(redis_client, command_name):
    """
        Retrieves command info from Redis
        Returns ModuleCommand
    """
    command_info = redis_client.execute_command("COMMAND INFO {}".format(command_name))
    # 1) 1) "graph.ADDEDGE"
    #    2) (integer) -1
    #    3) 1) write
    #    4) (integer) 1
    #    5) (integer) 1
    #    6) (integer) 1

    if len(command_info) != 1:
        return None

    command_info = command_info[0]

    command_name = command_info[0]
    command_arity = command_info[1]
    flags = command_info[2]
    first_key = command_info[3]
    last_key = command_info[4]
    step = command_info[5]

    return ModuleCommand(command_name, command_arity, flags, first_key, last_key, step)

def discover_modules_commands(path_to_module, module_args, redis_extra_args=None):
    """
        Retrieves module command(s) info.
        :param path_to_module: where does the module file is located
        :param module_args: command line arguments for the module
        :param redis_extra_args: command line arguments for redis
        Returns Module object populated with command(s) info.
    """
    with redis(redis_extra_args) as redis_client:
        core_redis_commands = _get_redis_commands(redis_client)
        module = _load_module(redis_client, path_to_module, module_args)
        if module is None:
            raise Exception("Failed to load module {} {}".format(path_to_module, module_args))

        extended_redis_commands = _get_redis_commands(redis_client)
        module_commands = (set(extended_redis_commands.keys()).symmetric_difference(set(core_redis_commands.keys())))
        # module_commands = extended_redis_commands - core_redis_commands

        for module_command in module_commands:
            command = _get_redis_command_info(redis_client, module_command)
            if command is None:
                raise Exception("Failed to retreive command info for {}".format(module_command))

            module.add_command(command)

        return module



import subprocess
import socket
import tempfile
import redis
import time
import os
import itertools
import sys

from ..common import *

# Environment variable pointing to the redis executable
REDIS_PATH_ENVVAR = 'REDIS_PATH'


def get_random_port():
    sock = socket.socket()
    sock.listen(0)
    _, port = sock.getsockname()
    sock.close()

    return port


class DisposableRedis(object):
    def __init__(self, port=None, path='redis-server', verbose=False, **extra_args):
        """
        :param port: port number to start the redis server on. Specify none to automatically generate
        :type port: int|None
        :param extra_args: any extra arguments kwargs will be passed to redis server as --key val
        """

        self._port = port
        self.verbose = verbose

        # this will hold the actual port the redis is listening on. It's equal to `_port` unless `_port` is None
        # in that case `port` is randomly generated
        self.port = None
        self.extra_args = list(itertools.chain(
                *(('--%s'%k, v) for k, v in extra_args.items())
               ))
        self.path = os.getenv(REDIS_PATH_ENVVAR, path)

    def _getRedisVersion(self):
        options = {
            'stderr': subprocess.PIPE,
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
        }
        p = subprocess.Popen(args=[self.path, '--version'], **options)
        while p.poll() is None:
            time.sleep(0.1)
        exit_code = p.poll()
        if exit_code != 0:
            raise Exception('Could not extract Redis version')
        out, err = p.communicate()
        out = out.decode('utf-8')
        v = out[out.find("v=") + 2:out.find("sha=") - 1].split('.')
        return int(v[0]) * 10000 + int(v[1]) * 100 + int(v[2])

    def __enter__(self):
        if self._port is None:
            self.port = get_random_port()
        else:
            self.port = self._port
        args = [self.path,
                '--port', str(self.port),
                '--dir', tempfile.gettempdir(),
                '--save', ''] + self.extra_args

        if self._getRedisVersion() >= 70000:
            args += ['--enable-module-command', 'yes']

        if self.verbose:
            out = sys.stdout
            err = sys.stderr
        else:
            out  = open(os.devnull, 'w')
            err = subprocess.STDOUT
        self.process = subprocess.Popen(
            args,
            #cwd=os.getcwd(),
            stdin=subprocess.PIPE,
            stdout=out,
            stderr=err,
            env=os.environ.copy(),
        )

        while True:
            try:
                self.client().ping()
                break
            except redis.ConnectionError:
                self.process.poll()
                if self.process.returncode is not None:
                    raise RuntimeError("Process has exited")
                time.sleep(0.1)

        return self.client()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.process.terminate()

    def client(self):
        """
        :rtype: redis.StrictRedis
        """

        return redis.StrictRedis(port=self.port, decode_responses=True)
